#!/usr/bin/perl

#------------------------------------------------------------------------------
# Program name: fsgui.pl
# Author      : Lance Geiger, lance@lancegeiger.com
# Revision    : 1.0M
# Last Updated: May 5, 2002
#
#
# Description
#
#   This is a graphical front-end for flexsim/IRflexsim network simulator.
#   It was written to be as modular as possible.  This program makes the
#   flexsim simulator more user friendly by grouping related flexsim 
#   command line options together, replacing numerical option with more
#   descriptive text ones, and verifying that legal values have been
#   entered.  The program also provides batch running capability, allowing
#   the user to specify a range of values at once for multiple flexsim
#   runs.
#
#------------------------------------------------------------------------------



# INCLUDES
use strict;		# Must declare variables, helps w/ debugging
use Tk;			# Allows use of Tk libraries (graphics)
use Cwd;		# Allows program to determine current directory


# GLOBAL VARIABLES
#
#   @options		array to store all options and their properties
#   @groupnames 	array of all group names found in ini file
#   %optiongroups	has storing the options contained within each group
#   @optionvalues	the current value of all options
#   %optionindexes	hash array where key=label and value=option ord #
#	$loaded			has an INI file been sucessfully loaded?
#   $mode			Simple or Advanced
#   $prog			flexsim or IRflexsim
#   LBOUND,UBOUND	upper and lower bound on supplied ord #'s from INI file
#   $batch_mode		yes or no
#   @batch_runs		array of command line strings to run

my (@options,@groupnames,%optiongroups,$loaded,@optionvalues,%optionindexes,$mode,$prog);
my ($LBOUND, $UBOUND,$batch_mode,@batch_runs);



# GLOBAL UI ELEMENTS
#
#   @groupframes	array of pointers to frames (one frame per group)
#	$main			pointer to main application window
#   $top			pointer to top level window frame
#   %frames			hash array storing pointers to frames for each field, accessed by field name
#   %labels			hash array storing pointers to text labels for each option
#   %helps			hash array storing pointers to help buttons for each option
#   $batchwin		pointer to window that displays list of submitted runs

my (@groupframes, $main, $top, %frames, %labels, %fields, %helps, $batchwin);



# USER CONFIGURABLE GLOBAL OPTIONS
#
#	$path_fs		path to the flexsim executable
#	$path_ir		path to the irflexsim executable
# 	$path_ini		path to the INI file
#	$output_dir		directory to store output files in

my ($path_fs, $path_ir, $path_ini, $output_dir);




# EXECUTION BEGINS HERE

print "Initializing ...\n";
&init;

print "Processing file ...\n";
&process_ini_file;

print "Displaying main window ...\n";
&display_main_window;

print "Cleaning up ...\n";
&clean_up;

# EXECUTION ENDS HERE



# PRIMARY FUNCTION DEFINITIONS

# init sets initial values for some of the global variables
sub init {

	$path_fs = "rsim";
	$path_ir = "irsim";
	$path_ini = "fsgui.ini";
	$output_dir = "output/";

	$mode = "simple";
	$prog = "flexsim";
	$batch_mode = "no";
}


# process_ini_file reads the fsgui config file and creates the hash to store every
# field and its properties
sub process_ini_file {

	my ($filename,$contents,@group_content,$count,$name,$data);
	
	# Initialize array bounds
	$UBOUND = 0;
	$LBOUND = 9e99;
	
	# Check if default file exists
	if (-e $path_ini) {

		# Open file and read it into $contents string then close file
		open(FILE1,$path_ini) || &fatal_error("Could not open configuration file");
		
		# Set flag to show INI file exists
		$loaded = 1;
		
		# Read each line into $contents variable
		while (<FILE1>) {

			# Ignore comments (lines that begin with a #)
			unless (/^#/ || /^[\l\r|\n]+/) {
				$contents = $contents . $_;
			}
		}
		close(FILE1);

		# Break sections into groups by searching for [GROUP NAME] tag
		@group_content = split(/[\s]*\[GROUP /, $contents);

		# Kill first set of group content since it's always blank (text before 1st tag)
		shift(@group_content);

		# Cycle through each set of group content and seperate group name from content
		foreach (@group_content) {
			($name,$data) = split(/\][\l\r|\n]+/,$_);
			unless ($name eq '') {

				# Add name of this group to global array for easy access
				push(@groupnames, $name);

				# Parse out options within group content
				&process_group($name,$data);
			}
		}
	
	# Else set flag if file does not exists
	} else {

		$loaded = 0;

	}
}


# display_main_window sets up the application window and begins the application loop
sub display_main_window {

	# Draw all currently visible fields	
	&create_window;
	
	# Enter Main Loop
	print "Entering main loop ...\n";
	MainLoop();
	
}


# clean_up closes files and performs necessary shut down work
sub clean_up {

	exit;

}


# SECONDARY FUNCTION DEFINITIONS

# process_group grabs the options listed within the group and stores them into the global
# array @options
sub process_group {
	my ($name,$data,@ops,$ord,$dummy,$option,$lab,$fs,$ir,$adv,$vis,$cmdline,$type,$def,$list,$val,$eff,$des);
	
	# Get parameters passed to this sub
	($name, $data) = @_;
	
	# Break group content into options by spliting at every newline
	@ops = split(/[\l\r|\n]+/,$data);
	
	# Get values within each option
	foreach (@ops) {
		
		# Grab description first since its in quotes
		#($dummy,$des,$dummy) = split(/"/,$_);
		
		# Remove white space from between fields
		s/,\s+(\S)/,$1/g;
		
		# Grab other values
		($ord,$option,$lab,$fs,$ir,$adv,$vis,$cmdline,$type,$def,$list,$val,$eff,$des) = split(/,/,$_);
		
		# Store into array indexed by option number
		$options[$ord] = "$option,$lab,$fs,$ir,$adv,$vis,$cmdline,$type,$def,$list,$val,$eff,$des";
		
		# Update UBOUND and LBOUND
		if ($ord < $LBOUND) {
			$LBOUND=$ord;
		}
		if ($ord > $UBOUND) {
			$UBOUND=$ord;
		}		
		
		# Associate this option with its correct group
		$optiongroups{"$name"} = $optiongroups{"$name"} . ":" . $ord;
		
		# Assign default value
		$optionvalues[$ord] = $def;
		
		# Create association with name
		$optionindexes{"$lab"} = $ord;
	}
}


# create_window sets up the main application window and all of its components initially
sub create_window {
	
	# Declare miscellaneous UI variables used throughout subroutine
	my ($group,@opt_list,$opt,$opt_name,$lab,$fs,$ir,$adv,$vis,$cmdline,$type,$def,$list,$val,$eff,$des,$name);
	my ($menu_bar, $file_menu, $help_menu, $control, $run_btn, $ref_btn);
	my ($prog_chooser, $prog_label, $mode_chooser, $mode_label, $batch_label, $batch_chooser);
	my ($spacer1, $spacer2, $spacer3, $bot, $status);
	my ($rowcount,$count,$par);
	
	# Make sure window hasn't already been created
	unless ($main) {

		# Create main window
		$main = MainWindow->new();
		$main->configure(-title=>'Flexsim Front End');
	
		# Create menu bar
		$menu_bar = $main->Frame(-relief=>'groove',-borderwidth=>3)->pack(-side=>'top',-fill=>'x');
	
		# Create File menu
		$file_menu = $menu_bar->Menubutton(-text=>'File')->pack(-side=>'left');
		$file_menu->command(-label=>'Config ...',-command=>sub{&config_window;});
		$file_menu->separator();
		$file_menu->command(-label=>'Exit',-command=>sub{$main->destroy});
	}

	
	# Only create rest of UI if INI file has been sucessfully loaded
	if ($loaded) {

		# Create main control panel
		$control = $main->Frame(-relief=>'ridge',-borderwidth=>3)->pack(-side=>'top',-fill=>'x');

		# Create flexsim/IRflexsim chooser
		$prog_label = $control->Label(-text=>"Program:")->pack(-side=>"left");	
		$prog_chooser = $control->Optionmenu(-options=>[qw(flexsim IRflexsim)],-command=>sub{ &refresh_window(); }, -variable=>\$prog)->pack(-side=>'left');

		# Add in blank label to space out control panel
		$spacer1 = $control->Label()->pack(-side=>"left",-padx=>15);

		# Create simple/adv chooser
		$mode_label = $control->Label(-text=>"Mode:")->pack(-side=>"left");	
		$mode_chooser = $control->Optionmenu(-options=>[qw(simple advanced)],-command=>sub{ &refresh_window(); },-variable=>\$mode)->pack(-side=>'left');

		# Add in blank label to space out control panel
		$spacer3 = $control->Label()->pack(-side=>"left",-padx=>15);

		# Create batch mode selector
		$batch_label = $control->Label(-text=>"Batch?")->pack(-side=>"left");	
		$batch_chooser = $control->Optionmenu(-options=>[qw(no yes)],-command=>sub{ &refresh_window(); },-variable=>\$batch_mode)->pack(-side=>'left');

		# Add in blank label to space out control panel
		$spacer2 = $control->Label()->pack(-side=>"left",-padx=>15);

		# Create run button
		$run_btn = $control->Button(-text=>'Run',-command=>sub{&run_program();})->pack(-side=>'left');

		# Create refresh button
		$ref_btn = $control->Button(-text=>'Refresh',-command=>sub{&refresh_window;})->pack(-side=>'left');

		# Create main frame and a child frame for each group
		$top = $main->Frame()->pack(-side=>'top',-fill=>'x');
		#$top = $main->Scrolled('Canvas',-scrollbars=>'se',-scrollregion=>[0,5000,0,5000])->pack(-side=>'top',-fill=>'x');

		# Initialize layout
		$rowcount=0;
		$count=0;
		$groupframes[$rowcount] = $top->Frame()->pack(-side=>'top',-anchor=>'n',-padx=>5, -pady=>5);
		$par = $groupframes[$rowcount];

		# Create labels and fields for each option, each within the proper group
		foreach $group (@groupnames) {

			# Make frame for this group and add label at top

			# We wish to have 3 groups per row, so after 3rd group is displayed, create a new frame
			# so next set of groups will be displayed on new row
			$count++;
			if ($count==4) {
				$rowcount++;
				$count=0;
				$groupframes[$rowcount] = $top->Frame()->pack(-side=>'top',-anchor=>'n',-padx=>5, -pady=>5);
				$par = $groupframes[$rowcount];
			}		

			# Create the frame for this group with a beveled border and title at the top center
			$frames{"$group"} = $par->Frame(-relief=>'groove',-borderwidth=>2)->pack(-side=>'left',-anchor=>'n',-padx=>5,-pady=>5);
			$labels{"$group"} = $frames{"$group"}->Label(-text=>$group,-fg=>'red')->pack(-side=>'top');

			# Make frame for labels, then one for entry boxes, then one for help buttons
			$frames{"lab_" . "$group"} = $frames{"$group"}->Frame()->pack(-side=>'left',-padx=>2, -pady=>2);
			$frames{"entry_" . "$group"} = $frames{"$group"}->Frame()->pack(-side=>'left',-padx=>2, -pady=>2);
			$frames{"help_" . "$group"} = $frames{"$group"}->Frame()->pack(-side=>'left',-padx=>2, -pady=>2);

			# Get list of options contained in this group
			@opt_list = split(/:/, $optiongroups{"$group"});

			# First option is always blank so kill it
			shift(@opt_list);

			# Display label, field, and help button for each option
			foreach $opt (@opt_list) {

				# Parse list of attributes for this option
				my ($opt_name,$lab,$fs,$ir,$adv,$vis,$cmdline,$type,$def,$list,$val,$eff,$des) = split(/,/,@options[$opt]);

				# Only display option if it is visible
				if ($vis==1) {

					# Make sure advanced options are only shown if in advanced mode
					unless ($mode eq "simple" && $adv==1) {

						# Make sure flexsim options are only shown for flexsim and same for IRflexsim
						unless ( ($prog eq "flexsim" && $fs==0) || ($prog eq "IRflexsim" && $ir==0) ) {

							# Display label for this option using $lab variable
							$labels{"$lab"} = $frames{"lab_" . "$group"}->Label(-text=>"$lab:",-pady=>7)->pack(-side=>'top',-anchor=>'e');

							# Create proper field type for this option, then insert default value ($def)
							if ($type eq "list") {

								# Display drop down list	
								$fields{"$lab"} = $frames{"entry_" . "$group"}->Optionmenu(-options=>[split(/;/,$list)],-variable=>\$optionvalues[$opt],-command=>\&refresh_window)->pack(-side=>'top',-pady=>4);
								$fields{"$lab"}->bind('Optionmenu','<ButtonPress>' => \&refresh_window);

							} elsif ($type eq "checkbox") {

								# Display checkbox
								$fields{"$lab"} = $frames{"entry_" . "$group"}->Checkbutton(-variable=>\$optionvalues[$opt],-command=>\&refresh_window)->pack(-side=>'top',-pady=>4);

							} else {

								# Display plain old text entry field
								$fields{"$lab"} = $frames{"entry_" . "$group"}->Entry(-width=>8,-justify=>'center',-textvariable=>\$optionvalues[$opt])->pack(-side=>'top',-pady=>4);

							}

							# Create help button for this option
							$helps{"$lab"} = $frames{"help_" . "$group"}->Button(-text=>'?',-command=>sub{&help_window($opt_name,$lab,$des)})->pack(-side=>'top');
						}
					}
				}
			}
		}

	} else {
	
		# Define minimum size for window so when no INI file is loaded the window does not
		# shrink into an unreadable size
		$main->minsize(qw(700 400));
		
	}
}


# refresh_window redraws the list of options reflecting a change in filtering or effects
sub refresh_window {

	# Only process if INI file loaded
	if ($loaded) {
		my ($i,$key,$group,$opt,@opt_list,$rowcount,$count,$par,$dummy);

		# Unpack and destory all objects from old window

		foreach $key (keys(%frames)) {
			$frames{$key}->packForget();
			delete $frames{$key};
		}	
		foreach $key (keys(%labels)) {
			$labels{$key}->packForget();
			delete $labels{$key};	
		}	
		foreach $key (keys(%fields)) {
			$fields{$key}->packForget();
			delete $fields{$key};	
		}	
		foreach $key (keys(%helps)) {
			$helps{$key}->packForget();
			delete $helps{$key};	
		}	
		foreach $key (@groupframes) {
			$key->packForget();
			$key = "";
		}

		# Initialize layout
		$rowcount=0;
		$count=0;
		$groupframes[$rowcount] = $top->Frame()->pack(-side=>'top',-anchor=>'n',-padx=>5, -pady=>5);
		$par = $groupframes[$rowcount];

		# Check validation rules of each option
		&process_validation;

		# Process inter-field effects
		&process_effects;

		# Create labels and fields for each option, each within the proper group
		foreach $group (@groupnames) {

			
			# Display 3 groups per row
			$count++;
			if ($count==4) {
				$rowcount++;
				$count=0;
				$groupframes[$rowcount] = $top->Frame()->pack(-side=>'top',-anchor=>'n',-padx=>5, -pady=>5);
				$par = $groupframes[$rowcount];
			}		

			# Make frame for this group and add label at top
			$frames{"$group"} = $par->Frame(-relief=>'groove',-borderwidth=>2)->pack(-side=>'left',-anchor=>'n',-padx=>5,-pady=>5);
			$labels{"$group"} = $frames{"$group"}->Label(-text=>$group,-fg=>'red')->pack(-side=>'top');

			# Make frame for labels, then one for entry boxes, then one for help buttons
			$frames{"lab_" . "$group"} = $frames{"$group"}->Frame()->pack(-side=>'left',-padx=>2, -pady=>2);
			$frames{"entry_" . "$group"} = $frames{"$group"}->Frame()->pack(-side=>'left',-padx=>2, -pady=>2);
			$frames{"help_" . "$group"} = $frames{"$group"}->Frame()->pack(-side=>'left',-padx=>2, -pady=>2);

			# Get list of options contained in this group
			@opt_list = split(/:/, $optiongroups{"$group"});

			# First option is always blank so kill it
			shift(@opt_list);

			# Display label, field, and help button for each option
			foreach $opt (@opt_list) {

				# Parse list of attributes for this option
				my ($opt_name,$lab,$fs,$ir,$adv,$vis,$cmdline,$type,$def,$list,$val,$eff,$des) = split(/,/,@options[$opt]);

				# Only display option if it is visible
				if ($vis==1) {

					# Make sure advanced options are only shown if in advanced mode
					unless ($mode eq "simple" && $adv==1) {

						# Make sure flexsim options are only shown for flexsim and same for IRflexsim
						unless ( ($prog eq "flexsim" && $fs==0) || ($prog eq "IRflexsim" && $ir==0) ) {

							# Display label for this option using $lab variable
							$labels{"$lab"} = $frames{"lab_" . "$group"}->Label(-text=>"$lab:",-pady=>7)->pack(-side=>'top',-anchor=>'e');

							# Create proper field type for this option, then insert default value ($def)

							if ($type eq "list") {

								# Make selected item the first thing
								my (@menuitems) = split(/;/,$list);
								my ($i,$num);
								$num = scalar(@menuitems);
								for ($i=0; $i<$num; $i++) {
									if ($menuitems[$i] eq $optionvalues[$opt]) {
										splice(@menuitems,$i,1);
										$num--;
									}
								}
								unshift(@menuitems,$optionvalues[$opt]);
								$list = join(';',@menuitems);

								# Display drop down list	
								$fields{"$lab"} = $frames{"entry_" . "$group"}->Optionmenu(-options=>[split(/;/,$list)],-variable=>\$optionvalues[$opt],-command=>\&refresh_window)->pack(-side=>'top',-pady=>4);

							} elsif ($type eq "checkbox") {

								# Display checkbox
								$fields{"$lab"} = $frames{"entry_" . "$group"}->Checkbutton(-variable=>\$optionvalues[$opt],-command=>\&refresh_window)->pack(-side=>'top',-pady=>4);

							} else {

								# Display plain old text entry field
								$fields{"$lab"} = $frames{"entry_" . "$group"}->Entry(-width=>8,-justify=>'center',-textvariable=>\$optionvalues[$opt])->pack(-side=>'top',-pady=>4);

							}

							# Create help button for this option
							$helps{"$lab"} = $frames{"help_" . "$group"}->Button(-text=>'?',-command=>sub{&help_window($opt_name,$lab,$des)})->pack(-side=>'top');
						}
					}
				} 
			}
		}
	}
}


# process_effects runs through the entire list of effects for every field and processes them
sub process_effects {

	my ($item,$effect,$res);

	# Cycle through each option
	foreach $item (@options) {
	
		my ($opt_name,$lab,$fs,$ir,$adv,$vis,$cmdline,$type,$def,$list,$val,$eff,$des) = split(/,/,$item);
			
		# Get list of effects (delimited by semicolon)
		my (@effects) = split(/;/,$eff);
		
		# Cycle through each effect for this option
		foreach $effect (@effects) {
		
			# Get condition, results of this effect
			my ($cond, @results) = split(/:/, $effect);
			
			# Parse the condition expression
			$cond = &parse_string($cond, $optionvalues[$optionindexes{"$lab"}]);
			
			# If the condition evaluates to TRUE, process results
			if (eval($cond)) {
						
				foreach $res (@results) {
				
					# Check for usage of SHOW()
					if ($res =~ /SHOW\(\[(\w+)\]\)/) {
						my ($opt_name,$lab,$fs,$ir,$adv,$vis,$cmdline,$type,$def,$list,$val,$eff,$des) = split(/,/,$options[$optionindexes{"$1"}]);
						$vis = 1;
						$options[$optionindexes{"$1"}] = join(',',$opt_name,$lab,$fs,$ir,$adv,$vis,$cmdline,$type,$def,$list,$val,$eff,$des);
				
					# Check for usage of HIDE()
					} elsif ($res =~ /HIDE\(\[(\w+)\]\)/) {
						my ($opt_name,$lab,$fs,$ir,$adv,$vis,$cmdline,$type,$def,$list,$val,$eff,$des) = split(/,/,$options[$optionindexes{"$1"}]);
						$vis = 0;
						$options[$optionindexes{"$1"}] = join(',',$opt_name,$lab,$fs,$ir,$adv,$vis,$cmdline,$type,$def,$list,$val,$eff,$des);
					
					# Check for usage of ENABLE()
					} elsif ($res =~ /ENABLE\(\[(\w+)\]\)/) {
						my ($opt_name,$lab,$fs,$ir,$adv,$vis,$cmdline,$type,$def,$list,$val,$eff,$des) = split(/,/,$options[$optionindexes{"$1"}]);
						$cmdline = 1;
						$options[$optionindexes{"$1"}] = join(',',$opt_name,$lab,$fs,$ir,$adv,$vis,$cmdline,$type,$def,$list,$val,$eff,$des);
					
					# Check for usage of DISABLE()
					} elsif ($res =~ /DISABLE\(\[(\w+)\]\)/) {
						my ($opt_name,$lab,$fs,$ir,$adv,$vis,$cmdline,$type,$def,$list,$val,$eff,$des) = split(/,/,$options[$optionindexes{"$1"}]);
						$cmdline = 2;
						$options[$optionindexes{"$1"}] = join(',',$opt_name,$lab,$fs,$ir,$adv,$vis,$cmdline,$type,$def,$list,$val,$eff,$des);
						
					# Else just parse and evaluate it
					} else {
						$res = &parse_string($res,$optionvalues[$optionindexes{$lab}]);
						eval($res);
					}
				}
			}
			
		}
	}

}


# process_validation runs through the entire list of validation rules for every field and checks them
sub process_validation {

	my ($item,$valid,$res,$value);

	# Cycle through each option
	foreach $item (@options) {
	
		my ($opt_name,$lab,$fs,$ir,$adv,$vis,$cmdline,$type,$def,$list,$val,$eff,$des) = split(/,/,$item);
		
		# Grab current value for easy access
		$value = $optionvalues[$optionindexes{"$lab"}];

		# Only validate if option is non-null
		if (length($value)>0) {
					
			# First look at option type and perform approiate validation
			if ($type eq 'count') {
				unless ($value>0 && int($value+0.999)==$value) {
					&error_window($lab,"Value must be an integer > 0");
				}

			} elsif ($type eq 'whole') {
				unless ($value>=0 && int($value+0.999)==$value) {
					&error_window($lab,"Value must be an integer >= 0");
				}

			} elsif ($type eq 'integer') {	
				unless (int($value+0.999)==$value) {
					&error_window($lab,"Value must be an integer");
				}

			} elsif ($type eq 'real') {
				unless ($value == $value+1-1) {
					# no checking for reals
				}

			} elsif ($type eq 'pow2') {
				unless ( int(log($value)/log(2)) == log($value)/log(2) ) {
					&error_window($lab,"Value must an integer power of 2");
				}

			} elsif ($type eq 'numstring') {	
				my ($flag, $chr, @digits);

				$flag = 0;
				@digits = split(//,$value);

				# Check each character, if not a digit than raise flag
				foreach $chr (@digits) {
					if ($chr != int($chr+0.999) || $chr == 0) {
						$flag = 1;
					}
				}

				if ($flag == 1) {
					&error_window($lab,"Value must be a string of non-zero integers");
				}
			}
		}
		
	
		# Get list of rules (delimited by semicolon)
		my (@rules) = split(/;/,$val);
		
		# Cycle through each effect for this option
		foreach $valid (@rules) {
		
			# Get condition, results of this effect
			my ($cond, @results) = split(/:/, $valid);
			
			# Parse the condition expression
			$cond = &parse_string($cond,$value);

			# If condition evaluates to TRUE, process results			
			if (eval($cond)) {
						
				foreach $res (@results) {
				
					my $res_parsed = &parse_string($res,$value);
					unless (eval($res_parsed)) {
						$res =~ s/\[SELF\]/$lab/ig;
						&error_window($lab,$res);
					}
			
				}
			}
			
		}
	}

}


# parse_string parses the psuedo-programming language used by fsgui and replaces references to fields with the actual value
sub parse_string {

	my ($i,$num,$value,@dummy);
	my ($str, $self) = @_;

	# Split by quote so we can ignore pieces between quotes	
	my (@strings) = split(/"/,$str);
	$num = scalar(@strings);

	# Process every other string in order skip pieces between quotes	
	for ($i=0; $i<$num; $i+=2) {

		# Check if this is an asignment operation, if so replace with proper name
		@dummy = split(/~=/,$strings[$i]);
		if (scalar(@dummy) > 1) {
		
			$dummy[0] =~ s/\[(\w+)\]/$1/;
			$strings[$i] = '$optionvalues[$optionindexes{' . $dummy[0] . '}]=' . &parse_string($dummy[1],$self);
		
		} else {

			# Replace [SELF] with actual value of the current field
			$strings[$i] =~ s/\[SELF\]/$self/g;
		
			# Replace [ALL] with 1
			$strings[$i] =~ s/\[ALL\]/1/g;

			# Replace [~DIAMETER] with n*(k/4) where k=average nodes per dimension and n=# of dimensions
			if ($strings[$i] =~ /\[~DIAMETER\]/) {
				my ($dia);
				$dia = $optionvalues[$optionindexes{'Dimension'}]*&numstring_avg($optionvalues[$optionindexes{'Size'}])/4;
				$strings[$i] =~ s/\[~DIAMETER\]/$dia/g;
			}
					
			# Replace [~NUMLINKS] with n*k^n where k=average nodes per dimension and n=# of dimensions
			if ($strings[$i] =~ /\[~NUMLINKS\]/) { 
				my ($numlinks);
				$numlinks = $optionvalues[$optionindexes{'Dimension'}]*&numstring_avg($optionvalues[$optionindexes{'Size'}])**$optionvalues[$optionindexes{'Dimension'}];
				$strings[$i] =~ s/\[~NUMLINKS\]/$numlinks/g;
			}
			
			# Replace [~NUMNODES] with number of nodes in network
			if ($strings[$i] =~ /\[~NUMNODES\]/) { 
				my ($numnodes);
				$numnodes = &numstring_prod($optionvalues[$optionindexes{'Size'}]);
				$strings[$i] =~ s/\[~NUMNODES\]/$numnodes/g;
			}
			
			# Replace [VAR] with actual value of the VAR option
			$strings[$i] =~ s/\[(\w+)\]/$optionvalues[$optionindexes{$1}]/g;
			
			# Handle use of POW2() function
			if ($strings[$i] =~ /POW2\((\S+)\)/) {
				$strings[$i] = pow2(eval($1));
			}
		}
	}

	# Join the newly processed string fragments back together
	$str = join('"', @strings);

	# Add a trailing quote if necessary	
	if ($num%2==0) {
		$str = $str . '"';
	}
	
	return $str;
}


# error_window alerts user when a validation rule of an option has been violated
sub error_window {
	
	# Create new window
	my $error_popup = MainWindow->new();
	$error_popup->configure(-title=>'Error');
	
	# Display option label
	my $cmd = $error_popup->Label(-text=>"$_[0]")->pack(-side=>'top', -pady=>5);
	
	# Create frame 
	my $box = $error_popup->Frame(-relief=>'groove',-borderwidth=>2)->pack(-side=>'top',-padx=>5,-pady=>5);
	
	# Display value entered, rule violated, and close button
	my $user = $box->Label(-text=>"You entered: " . $optionvalues[$optionindexes{$_[0]}])->pack(-side=>'top', pady=>5);
	my $err_msg = $box->Label(-text=>"Rule violated: " . $_[1])->pack(-side=>'top', pady=>5);
	my $btn = $box->Button(-text=>'Close',-command=>sub{$error_popup->destroy})->pack(-side=>'top');
}


# help_window creates a new "popup" window which displays the related description for the field it belongs to
sub help_window {
	
	# Create new window
	my $popup = MainWindow->new();
	$popup->configure(-title=>'Help');
	
	# Display option label and name in ()
	my $cmd = $popup->Label(-text=>"$_[1] ($_[0])")->pack(-side=>'top', -pady=>5);
	
	# Create frame 
	my $box = $popup->Frame(-relief=>'groove',-borderwidth=>2)->pack(-side=>'top',-padx=>5,-pady=>5);
	
	# Display option name, description. and close button
	my $msg = $box->Label(-text=>$_[2])->pack(-side=>'top', pady=>10);
	my $btn = $box->Button(-text=>'Close',-command=>sub{$popup->destroy})->pack(-side=>'top');

}


# config_window creates a window that allows used to configure a few of the program parameters
sub config_window {

	# Create new window
	my $config_win = MainWindow->new();
	$config_win->configure(-title=>'Config');
	
	# Display window title
	my $cmd = $config_win->Label(-text=>'Please specify the parameters below:')->pack(-side=>'top', -pady=>5);
	
	# Create frame for labels and one for fields
	my $top_frame = $config_win->Frame()->pack(-side=>'top');
	my $lab_frame = $top_frame->Frame()->pack(-side=>'left',-padx=>5,-pady=>5);
	my $fld_frame = $top_frame->Frame()->pack(-side=>'left',-padx=>5,-pady=>5);
	
	# Add labels
	my $lab_path_fs = $lab_frame->Label(-text=>'Path to flexsim executable:',-pady=>7)->pack(-side=>'top',-anchor=>'e');
	my $lab_path_ir = $lab_frame->Label(-text=>'Path to IRflexsim executable:',-pady=>7)->pack(-side=>'top',-anchor=>'e');
	my $lab_path_ini = $lab_frame->Label(-text=>'Path to INI file:',-pady=>7)->pack(-side=>'top',-anchor=>'e');
	my $lab_output_dir = $lab_frame->Label(-text=>'Output directory:',-pady=>7)->pack(-side=>'top',-anchor=>'e');
	
	# Display list of fields to edit 
	my $fields_path_fs = $fld_frame->Entry(-width=>50,-justify=>'center',-textvariable=>\$path_fs)->pack(-side=>'top',-pady=>4);
	my $fields_path_ir = $fld_frame->Entry(-width=>50,-justify=>'center',-textvariable=>\$path_ir)->pack(-side=>'top',-pady=>4);
	my $fields_path_ini = $fld_frame->Entry(-width=>50,-justify=>'center',-textvariable=>\$path_ini)->pack(-side=>'top',-pady=>4);
	my $fields_output_dir = $fld_frame->Entry(-width=>50,-justify=>'center',-textvariable=>\$output_dir)->pack(-side=>'top',-pady=>4);
		
	# Create another frame for button
	my $bot = $config_win->Frame()->pack(-side=>'top',-pady=>5);
	my $btn = $bot->Button(-text=>'Apply',-command=>sub{ $config_win->destroy; unless($loaded) { &process_ini_file; &create_window; } })->pack(-side=>'top');

}

# run_program gathers the current values of all the enabled flexsim parameters and passes these
# to the flexsim or IRflexsim program
sub run_program {

	# Process effects and check validity one last time
	&process_effects;
	&process_validation;

	my ($command_line,@lines,$opt,$item,$dummy);
	
	# Go through each option, if it's visible and not disabled then add it
	for ($opt=$LBOUND; $opt<=$UBOUND; $opt++) {
		my ($opt_name,$lab,$fs,$ir,$adv,$vis,$cmdline,$type,$def,$list,$val,$eff,$des) = split(/,/,@options[$opt]);		
		
		if ($vis==1 || $cmdline==1) {
			if (!($mode eq "simple" && $adv==1) || $cmdline==1) {
				if ( !(($prog eq "flexsim" && $fs==0) || ($prog eq "IRflexsim" && $ir==0)) || $cmdline==1) {
					if ($optionvalues[$opt] ne "" || $cmdline==1) {
						if ($cmdline!=2) {				
							if ($type eq "list") {

								($item,$dummy) = split(/=/,$optionvalues[$opt]);		
								$item =~ s/"//g;
								$command_line = $command_line . " " . $opt_name . "=" . $item;

							} else {

								$command_line = $command_line . " " . $opt_name . "=" . $optionvalues[$opt];
							}
						}
					}
				}
			}
		}
	}
	
	# Check output directory
	unless ($output_dir =~ /\/$/) {
		$output_dir = $output_dir . "/";
	}
	
	# Prepend program name to front
	if ($prog eq "flexsim") {
		$command_line = $path_fs . $command_line . " > output.txt";
	} else {
		$command_line = $path_ir . $command_line . " > output.txt";
	}
	
	# For now just display string and exit program
	if ($batch_mode eq "no") {
		
		# Get the directory that this program was run from
		my $first_dir = cwd;
	
		# Add trailing / if necessary
		unless ($first_dir =~ /\/$/) {
			unless ($output_dir =~ /^\//) {
				$first_dir = $first_dir . "/";
			}
		}
	
		# Prepend directory to command line so it can be run from output dir
		$command_line = $first_dir . $command_line;
	
		# Make output directory if necessary
		unless (-e $output_dir) {
			system("mkdir " . $output_dir);
		}
		
		# Go into directory
		chdir($output_dir);

		# Run in background
		$command_line = $command_line . " &";
			
		# Run the program
		print "$command_line\n";
		system($command_line);
		
		# All done so exit program
		exit;
	
	} else {
		my ($box);
		
		push(@batch_runs, $command_line);
		
		# Check if window already exists, if so destory it
		if ($batchwin) {
			$batchwin->destroy;
			$batchwin = undef;
		}
		
		# Create new window to list batch runs
		$batchwin = MainWindow->new();
		$batchwin->configure(-title=>'Runs to be submitted');
			
		# Display option label and name in ()
		my $cmd = $batchwin->Label(-text=>"The following command lines will be run:")->pack(-side=>'top', -pady=>5);
		
		# Create frame 
		$box = $batchwin->Frame(-relief=>'groove',-borderwidth=>2)->pack(-side=>'top',-padx=>5,-pady=>5);
		
		# Display list of command lines
		foreach $opt (@batch_runs) {
			
			# Truncate command line listing to 50 characters
			my $string = $opt;
			
			if (length($opt) > 100) {
				$string = substr($string,0,100) . " ...";
			}
			
			my $msg = $box->Label(-text=>$string)->pack(-side=>'top', pady=>1);
		}
		my $btn = $batchwin->Button(-text=>'Begin',-command=>sub{&batch_run_program; $batchwin->destroy; $batchwin=undef;})->pack(-side=>'top');
	}
}


# batch_run_program will send the entire list of batch runs to the shell, one by one
sub batch_run_program {

	my ($i,$num,$dirname);
	
	$num = scalar(@batch_runs);
	
	# Get the directory that this program was run from
	my $first_dir = cwd;

	# Add trailing / if necessary
	unless ($first_dir =~ /\/$/) {
		unless ($output_dir =~ /^\//) {
			$first_dir = $first_dir . "/";
		}
	}
	
	# Create root output directory if it doesn't already exist
	unless (-e $output_dir) {
		system("mkdir " . $output_dir);
	}

	for ($i=0; $i<$num; $i++) {
	
		# Determine output directory name for this run - make sure it is at least 6 characters
		$dirname = $i;
		while (length($dirname) < 6) {
			$dirname = '0' . $dirname;
		}
	
		# Append output_dir to name
		$dirname = $output_dir . $dirname;
	
		# Return to starting directory
		chdir($first_dir);

		# Make directory for this run
		unless (-e $dirname) {
			system("mkdir " . $dirname);
		}
		
		# Go into this new directory
		chdir($dirname);

		# Prepend directory to command line so it can be run from output dir
		$batch_runs[$i] = $first_dir . $batch_runs[$i];	
		
		# Run in background
		$batch_runs[$i] = $batch_runs[$i] . " &";
		
		# Execute run
		print "$batch_runs[$i]\n";
		system($batch_runs[$i]);
	}

	# All done so exit program
	exit;
}



# UTILITY FUNCTION DEFINITIONS


# fatal_error simply displays the passed message and kills the program
sub fatal_error {
	print $_[0] . "\n";
	#exit;
}


# pow2(x) returns the cloest power of 2 that is greater than or equal to x
sub pow2 {

	my ($i);
	my ($x) = $_[0];
	
	$i = 0;
	
	while(2**$i < $x) {
		$i++;
	}
	
	return 2**$i;

}


# numstring_prod returns the product (2^x) of the values represented by the numstring
sub numstring_prod {

	my ($i,@digits,$num,$prod);
	
	@digits = split(//,$_[0]);
	$num = scalar(@digits);
	$prod = 1;
	
	for ($i=0; $i<$num; $i++) {
		$prod *= 2**$digits[$i];
	}
	
	return $prod;
	
}


# numstring_avg returns the average value (2^x) represented by the digits in a numstring
sub numstring_avg {

	my ($i,@digits,$sum,$num);
	
	@digits = split(//,$_[0]);
	$num = scalar(@digits);
	$sum = 0;
	
	for ($i=0; $i<$num; $i++) {
		$sum += 2**$digits[$i];
	}
	
	return $sum/$num;

}
